@@ -1,18 +1,15 @@
 //! This module defines the `Type` type, representing the dynamic form of a component interface type.
 
 use crate::component::matching::InstanceType;
-use crate::component::values::{self, Val};
-use crate::{ExternType, FuncType};
-use anyhow::{anyhow, Result};
-use std::fmt;
-use std::mem;
-use std::ops::Deref;
-use std::sync::Arc;
+use crate::{Engine, ExternType, FuncType};
+use alloc::sync::Arc;
+use core::fmt;
+use core::ops::Deref;
 use wasmtime_environ::component::{
-    CanonicalAbiInfo, ComponentTypes, InterfaceType, ResourceIndex, TypeComponentIndex,
-    TypeComponentInstanceIndex, TypeDef, TypeEnumIndex, TypeFlagsIndex, TypeFuncIndex,
-    TypeListIndex, TypeModuleIndex, TypeOptionIndex, TypeRecordIndex, TypeResourceTableIndex,
-    TypeResultIndex, TypeTupleIndex, TypeVariantIndex,
+    ComponentTypes, InterfaceType, ResourceIndex, TypeComponentIndex, TypeComponentInstanceIndex,
+    TypeDef, TypeEnumIndex, TypeFlagsIndex, TypeFuncIndex, TypeListIndex, TypeModuleIndex,
+    TypeOptionIndex, TypeRecordIndex, TypeResourceTableIndex, TypeResultIndex, TypeTupleIndex,
+    TypeVariantIndex,
 };
 use wasmtime_environ::PrimaryMap;
 
@@ -183,18 +180,21 @@ impl TypeChecker<'_> {
         if a.cases.len() != b.cases.len() {
             return false;
         }
-        a.cases.iter().zip(b.cases.iter()).all(|(a_case, b_case)| {
-            if a_case.name != b_case.name {
-                return false;
-            }
-            match (a_case.ty, b_case.ty) {
-                (Some(a_case_ty), Some(b_case_ty)) => {
-                    self.interface_types_equal(a_case_ty, b_case_ty)
+        a.cases
+            .iter()
+            .zip(b.cases.iter())
+            .all(|((a_name, a_ty), (b_name, b_ty))| {
+                if a_name != b_name {
+                    return false;
                 }
-                (None, None) => true,
-                _ => false,
-            }
-        })
+                match (a_ty, b_ty) {
+                    (Some(a_case_ty), Some(b_case_ty)) => {
+                        self.interface_types_equal(*a_case_ty, *b_case_ty)
+                    }
+                    (None, None) => true,
+                    _ => false,
+                }
+            })
     }
 
     fn results_equal(&self, r1: TypeResultIndex, r2: TypeResultIndex) -> bool {
@@ -259,16 +259,11 @@ impl PartialEq for List {
 impl Eq for List {}
 
 impl List {
-    /// Instantiate this type with the specified `values`.
-    pub fn new_val(&self, values: Box<[Val]>) -> Result<Val> {
-        Ok(Val::List(values::List::new(self, values)?))
-    }
-
     pub(crate) fn from(index: TypeListIndex, ty: &InstanceType<'_>) -> Self {
         List(Handle::new(index, ty))
     }
 
-    /// Retreive the element type of this `list`.
+    /// Retrieve the element type of this `list`.
     pub fn ty(&self) -> Type {
         Type::from(&self.0.types[self.0.index].element, &self.0.instance())
     }
@@ -288,11 +283,6 @@ pub struct Field<'a> {
 pub struct Record(Handle<TypeRecordIndex>);
 
 impl Record {
-    /// Instantiate this type with the specified `values`.
-    pub fn new_val<'a>(&self, values: impl IntoIterator<Item = (&'a str, Val)>) -> Result<Val> {
-        Ok(Val::Record(values::Record::new(self, values)?))
-    }
-
     pub(crate) fn from(index: TypeRecordIndex, ty: &InstanceType<'_>) -> Self {
         Record(Handle::new(index, ty))
     }
@@ -319,11 +309,6 @@ impl Eq for Record {}
 pub struct Tuple(Handle<TypeTupleIndex>);
 
 impl Tuple {
-    /// Instantiate this type ith the specified `values`.
-    pub fn new_val(&self, values: Box<[Val]>) -> Result<Val> {
-        Ok(Val::Tuple(values::Tuple::new(self, values)?))
-    }
-
     pub(crate) fn from(index: TypeTupleIndex, ty: &InstanceType<'_>) -> Self {
         Tuple(Handle::new(index, ty))
     }
@@ -358,24 +343,19 @@ pub struct Case<'a> {
 pub struct Variant(Handle<TypeVariantIndex>);
 
 impl Variant {
-    /// Instantiate this type with the specified case `name` and `value`.
-    pub fn new_val(&self, name: &str, value: Option<Val>) -> Result<Val> {
-        Ok(Val::Variant(values::Variant::new(self, name, value)?))
-    }
-
     pub(crate) fn from(index: TypeVariantIndex, ty: &InstanceType<'_>) -> Self {
         Variant(Handle::new(index, ty))
     }
 
     /// Retrieve the cases of this `variant` in declaration order.
     pub fn cases(&self) -> impl ExactSizeIterator<Item = Case> {
-        self.0.types[self.0.index].cases.iter().map(|case| Case {
-            name: &case.name,
-            ty: case
-                .ty
-                .as_ref()
-                .map(|ty| Type::from(ty, &self.0.instance())),
-        })
+        self.0.types[self.0.index]
+            .cases
+            .iter()
+            .map(|(name, ty)| Case {
+                name: name,
+                ty: ty.as_ref().map(|ty| Type::from(ty, &self.0.instance())),
+            })
     }
 }
 
@@ -392,11 +372,6 @@ impl Eq for Variant {}
 pub struct Enum(Handle<TypeEnumIndex>);
 
 impl Enum {
-    /// Instantiate this type with the specified case `name`.
-    pub fn new_val(&self, name: &str) -> Result<Val> {
-        Ok(Val::Enum(values::Enum::new(self, name)?))
-    }
-
     pub(crate) fn from(index: TypeEnumIndex, ty: &InstanceType<'_>) -> Self {
         Enum(Handle::new(index, ty))
     }
@@ -423,11 +398,6 @@ impl Eq for Enum {}
 pub struct OptionType(Handle<TypeOptionIndex>);
 
 impl OptionType {
-    /// Instantiate this type with the specified `value`.
-    pub fn new_val(&self, value: Option<Val>) -> Result<Val> {
-        Ok(Val::Option(values::OptionVal::new(self, value)?))
-    }
-
     pub(crate) fn from(index: TypeOptionIndex, ty: &InstanceType<'_>) -> Self {
         OptionType(Handle::new(index, ty))
     }
@@ -451,16 +421,11 @@ impl Eq for OptionType {}
 pub struct ResultType(Handle<TypeResultIndex>);
 
 impl ResultType {
-    /// Instantiate this type with the specified `value`.
-    pub fn new_val(&self, value: Result<Option<Val>, Option<Val>>) -> Result<Val> {
-        Ok(Val::Result(values::ResultVal::new(self, value)?))
-    }
-
     pub(crate) fn from(index: TypeResultIndex, ty: &InstanceType<'_>) -> Self {
         ResultType(Handle::new(index, ty))
     }
 
-    /// Retrieve the `ok` type parameter for this `option`.
+    /// Retrieve the `ok` type parameter for this `result`.
     pub fn ok(&self) -> Option<Type> {
         Some(Type::from(
             self.0.types[self.0.index].ok.as_ref()?,
@@ -468,7 +433,7 @@ impl ResultType {
         ))
     }
 
-    /// Retrieve the `err` type parameter for this `option`.
+    /// Retrieve the `err` type parameter for this `result`.
     pub fn err(&self) -> Option<Type> {
         Some(Type::from(
             self.0.types[self.0.index].err.as_ref()?,
@@ -490,11 +455,6 @@ impl Eq for ResultType {}
 pub struct Flags(Handle<TypeFlagsIndex>);
 
 impl Flags {
-    /// Instantiate this type with the specified flag `names`.
-    pub fn new_val(&self, names: &[&str]) -> Result<Val> {
-        Ok(Val::Flags(values::Flags::new(self, names)?))
-    }
-
     pub(crate) fn from(index: TypeFlagsIndex, ty: &InstanceType<'_>) -> Self {
         Flags(Handle::new(index, ty))
     }
@@ -506,10 +466,6 @@ impl Flags {
             .iter()
             .map(|name| name.deref())
     }
-
-    pub(crate) fn canonical_abi(&self) -> &CanonicalAbiInfo {
-        &self.0.types[self.0.index].abi
-    }
 }
 
 impl PartialEq for Flags {
@@ -678,31 +634,6 @@ impl Type {
         }
     }
 
-    /// Checks whether type of `value` is a subtype of `self`.
-    ///
-    /// # Errors
-    ///
-    /// Returns an error in case of a type mismatch
-    pub(crate) fn is_supertype_of(&self, value: &Val) -> Result<()> {
-        let other = &value.ty();
-        if self == other
-            || matches!((self, other), (Self::Borrow(s), Self::Own(other)) if s == other)
-        {
-            Ok(())
-        } else if mem::discriminant(self) != mem::discriminant(other) {
-            Err(anyhow!(
-                "type mismatch: expected {}, got {}",
-                self.desc(),
-                other.desc()
-            ))
-        } else {
-            Err(anyhow!(
-                "type mismatch for {}, possibly due to mixing distinct composite types",
-                self.desc()
-            ))
-        }
-    }
-
     /// Convert the specified `InterfaceType` to a `Type`.
     pub(crate) fn from(ty: &InterfaceType, instance: &InstanceType<'_>) -> Self {
         match ty {
@@ -799,24 +730,30 @@ impl Module {
     }
 
     /// Iterates over imports of the module
-    pub fn imports(&self) -> impl ExactSizeIterator<Item = ((&str, &str), ExternType)> {
+    pub fn imports<'a>(
+        &'a self,
+        engine: &'a Engine,
+    ) -> impl ExactSizeIterator<Item = ((&str, &str), ExternType)> + 'a {
         self.0.types[self.0.index]
             .imports
             .iter()
             .map(|((namespace, name), ty)| {
                 (
                     (namespace.as_str(), name.as_str()),
-                    ExternType::from_wasmtime(self.0.types.module_types(), ty),
+                    ExternType::from_wasmtime(engine, self.0.types.module_types(), ty),
                 )
             })
     }
 
     /// Iterates over exports of the module
-    pub fn exports(&self) -> impl ExactSizeIterator<Item = (&str, ExternType)> {
+    pub fn exports<'a>(
+        &'a self,
+        engine: &'a Engine,
+    ) -> impl ExactSizeIterator<Item = (&str, ExternType)> + 'a {
         self.0.types[self.0.index].exports.iter().map(|(name, ty)| {
             (
                 name.as_str(),
-                ExternType::from_wasmtime(self.0.types.module_types(), ty),
+                ExternType::from_wasmtime(engine, self.0.types.module_types(), ty),
             )
         })
     }
@@ -832,35 +769,45 @@ impl Component {
     }
 
     /// Returns import associated with `name`, if such exists in the component
-    pub fn get_import(&self, name: &str) -> Option<ComponentItem> {
+    pub fn get_import(&self, engine: &Engine, name: &str) -> Option<ComponentItem> {
         self.0.types[self.0.index]
             .imports
             .get(name)
-            .map(|ty| ComponentItem::from(ty, &self.0.instance()))
+            .map(|ty| ComponentItem::from(engine, ty, &self.0.instance()))
     }
 
     /// Iterates over imports of the component
-    pub fn imports(&self) -> impl ExactSizeIterator<Item = (&str, ComponentItem)> {
-        self.0.types[self.0.index]
-            .imports
-            .iter()
-            .map(|(name, ty)| (name.as_str(), ComponentItem::from(ty, &self.0.instance())))
+    pub fn imports<'a>(
+        &'a self,
+        engine: &'a Engine,
+    ) -> impl ExactSizeIterator<Item = (&str, ComponentItem)> + 'a {
+        self.0.types[self.0.index].imports.iter().map(|(name, ty)| {
+            (
+                name.as_str(),
+                ComponentItem::from(engine, ty, &self.0.instance()),
+            )
+        })
     }
 
     /// Returns export associated with `name`, if such exists in the component
-    pub fn get_export(&self, name: &str) -> Option<ComponentItem> {
+    pub fn get_export(&self, engine: &Engine, name: &str) -> Option<ComponentItem> {
         self.0.types[self.0.index]
             .exports
             .get(name)
-            .map(|ty| ComponentItem::from(ty, &self.0.instance()))
+            .map(|ty| ComponentItem::from(engine, ty, &self.0.instance()))
     }
 
     /// Iterates over exports of the component
-    pub fn exports(&self) -> impl ExactSizeIterator<Item = (&str, ComponentItem)> {
-        self.0.types[self.0.index]
-            .exports
-            .iter()
-            .map(|(name, ty)| (name.as_str(), ComponentItem::from(ty, &self.0.instance())))
+    pub fn exports<'a>(
+        &'a self,
+        engine: &'a Engine,
+    ) -> impl ExactSizeIterator<Item = (&str, ComponentItem)> + 'a {
+        self.0.types[self.0.index].exports.iter().map(|(name, ty)| {
+            (
+                name.as_str(),
+                ComponentItem::from(engine, ty, &self.0.instance()),
+            )
+        })
     }
 }
 
@@ -874,19 +821,24 @@ impl ComponentInstance {
     }
 
     /// Returns export associated with `name`, if such exists in the component instance
-    pub fn get_export(&self, name: &str) -> Option<ComponentItem> {
+    pub fn get_export(&self, engine: &Engine, name: &str) -> Option<ComponentItem> {
         self.0.types[self.0.index]
             .exports
             .get(name)
-            .map(|ty| ComponentItem::from(ty, &self.0.instance()))
+            .map(|ty| ComponentItem::from(engine, ty, &self.0.instance()))
     }
 
     /// Iterates over exports of the component instance
-    pub fn exports(&self) -> impl ExactSizeIterator<Item = (&str, ComponentItem)> {
-        self.0.types[self.0.index]
-            .exports
-            .iter()
-            .map(|(name, ty)| (name.as_str(), ComponentItem::from(ty, &self.0.instance())))
+    pub fn exports<'a>(
+        &'a self,
+        engine: &'a Engine,
+    ) -> impl ExactSizeIterator<Item = (&str, ComponentItem)> {
+        self.0.types[self.0.index].exports.iter().map(|(name, ty)| {
+            (
+                name.as_str(),
+                ComponentItem::from(engine, ty, &self.0.instance()),
+            )
+        })
     }
 }
 
@@ -910,7 +862,7 @@ pub enum ComponentItem {
 }
 
 impl ComponentItem {
-    pub(crate) fn from(def: &TypeDef, ty: &InstanceType<'_>) -> Self {
+    pub(crate) fn from(engine: &Engine, def: &TypeDef, ty: &InstanceType<'_>) -> Self {
         match def {
             TypeDef::Component(idx) => Self::Component(Component::from(*idx, ty)),
             TypeDef::ComponentInstance(idx) => {
@@ -920,9 +872,26 @@ impl ComponentItem {
             TypeDef::Interface(iface_ty) => Self::Type(Type::from(iface_ty, ty)),
             TypeDef::Module(idx) => Self::Module(Module::from(*idx, ty)),
             TypeDef::CoreFunc(idx) => {
-                Self::CoreFunc(FuncType::from_wasm_func_type(ty.types[*idx].clone()))
+                let subty = &ty.types[*idx];
+                Self::CoreFunc(FuncType::from_wasm_func_type(
+                    engine,
+                    subty.is_final,
+                    subty.supertype,
+                    subty.unwrap_func().clone(),
+                ))
+            }
+            TypeDef::Resource(idx) => {
+                let resource_index = ty.types[*idx].ty;
+                let ty = match ty.resources.get(resource_index) {
+                    // This resource type was substituted by a linker for
+                    // example so it's replaced here.
+                    Some(ty) => *ty,
+
+                    // This resource type was not substituted.
+                    None => ResourceType::uninstantiated(&ty.types, resource_index),
+                };
+                Self::Resource(ty)
             }
-            TypeDef::Resource(idx) => Self::Resource(ty.resources[ty.types[*idx].ty]),
         }
     }
 }
