@@ -1,19 +1,20 @@
 use crate::func::HostFunc;
 use crate::instance::InstancePre;
 use crate::store::StoreOpaque;
+use crate::{prelude::*, IntoFunc};
 use crate::{
     AsContext, AsContextMut, Caller, Engine, Extern, ExternType, Func, FuncType, ImportType,
-    Instance, IntoFunc, Module, StoreContextMut, Val, ValRaw, ValType,
+    Instance, Module, StoreContextMut, Val, ValRaw, ValType, WasmTyList,
 };
-use anyhow::{bail, Context, Result};
+use alloc::sync::Arc;
+use core::fmt;
+#[cfg(feature = "async")]
+use core::future::Future;
+use core::marker;
+#[cfg(feature = "async")]
+use core::pin::Pin;
+use hashbrown::hash_map::{Entry, HashMap};
 use log::warn;
-use std::collections::hash_map::{Entry, HashMap};
-#[cfg(feature = "async")]
-use std::future::Future;
-use std::marker;
-#[cfg(feature = "async")]
-use std::pin::Pin;
-use std::sync::Arc;
 
 /// Structure used to link wasm modules/instances together.
 ///
@@ -65,7 +66,7 @@ use std::sync::Arc;
 /// [`Linker`] value at program start up and use that continuously for each
 /// [`Store`] created over the lifetime of the program.
 ///
-/// Note that once [`Store`]-owned items, such as [`Global`], are defined witin
+/// Note that once [`Store`]-owned items, such as [`Global`], are defined within
 /// a [`Linker`] then it is no longer compatible with any [`Store`]. At that
 /// point only the [`Store`] that owns the [`Global`] can be used to instantiate
 /// modules.
@@ -120,9 +121,9 @@ pub(crate) enum Definition {
 /// This is a sort of slimmed down `ExternType` which notably doesn't have a
 /// `FuncType`, which is an allocation, and additionally retains the current
 /// size of the table/memory.
-#[derive(Clone)]
+#[derive(Clone, Debug)]
 pub(crate) enum DefinitionType {
-    Func(wasmtime_runtime::VMSharedTypeIndex),
+    Func(wasmtime_environ::VMSharedTypeIndex),
     Global(wasmtime_environ::Global),
     // Note that tables and memories store not only the original type
     // information but additionally the current size of the table/memory, as
@@ -132,45 +133,6 @@ pub(crate) enum DefinitionType {
     Memory(wasmtime_environ::Memory, u64),
 }
 
-macro_rules! generate_wrap_async_func {
-    ($num:tt $($args:ident)*) => (paste::paste!{
-        /// Asynchronous analog of [`Linker::func_wrap`].
-        ///
-        /// For more information also see
-        /// [`Func::wrapN_async`](crate::Func::wrap1_async).
-        #[allow(non_snake_case)]
-        #[cfg(feature = "async")]
-        #[cfg_attr(nightlydoc, doc(cfg(feature = "async")))]
-        pub fn [<func_wrap $num _async>]<$($args,)* R>(
-            &mut self,
-            module: &str,
-            name: &str,
-            func: impl for<'a> Fn(Caller<'a, T>, $($args),*) -> Box<dyn Future<Output = R> + Send + 'a> + Send + Sync + 'static,
-        ) -> Result<&mut Self>
-        where
-            $($args: crate::WasmTy,)*
-            R: crate::WasmRet,
-        {
-            assert!(
-                self.engine.config().async_support,
-                concat!(
-                    "cannot use `func_wrap",
-                    $num,
-                    "_async` without enabling async support on the config",
-                ),
-            );
-            self.func_wrap(module, name, move |mut caller: Caller<'_, T>, $($args: $args),*| {
-                let async_cx = caller.store.as_context_mut().0.async_cx().expect("Attempt to start async function on dying fiber");
-                let mut future = Pin::from(func(caller, $($args),*));
-                match unsafe { async_cx.block_on(future.as_mut()) } {
-                    Ok(ret) => ret.into_fallible(),
-                    Err(e) => R::fallible_from_error(e),
-                }
-            })
-        }
-    })
-}
-
 impl<T> Linker<T> {
     /// Creates a new [`Linker`].
     ///
@@ -255,8 +217,7 @@ impl<T> Linker<T> {
     /// Implement any imports of the given [`Module`] with a function which traps.
     ///
     /// By default a [`Linker`] will error when unknown imports are encountered
-    /// in a command module while using [`Linker::module`]. Use this function
-    /// when
+    /// in a command module while using [`Linker::module`].
     ///
     /// This method can be used to allow unknown imports from command modules.
     ///
@@ -274,8 +235,6 @@ impl<T> Linker<T> {
     /// # Ok(())
     /// # }
     /// ```
-    #[cfg(any(feature = "cranelift", feature = "winch"))]
-    #[cfg_attr(nightlydoc, doc(cfg(any(feature = "cranelift", feature = "winch"))))]
     pub fn define_unknown_imports_as_traps(&mut self, module: &Module) -> anyhow::Result<()> {
         for import in module.imports() {
             if let Err(import_err) = self._get_by_import(&import) {
@@ -310,8 +269,6 @@ impl<T> Linker<T> {
     /// # Ok(())
     /// # }
     /// ```
-    #[cfg(any(feature = "cranelift", feature = "winch"))]
-    #[cfg_attr(nightlydoc, doc(cfg(any(feature = "cranelift", feature = "winch"))))]
     pub fn define_unknown_imports_as_default_values(
         &mut self,
         module: &Module,
@@ -320,6 +277,13 @@ impl<T> Linker<T> {
             if let Err(import_err) = self._get_by_import(&import) {
                 if let ExternType::Func(func_ty) = import_err.ty() {
                     let result_tys: Vec<_> = func_ty.results().collect();
+
+                    for ty in &result_tys {
+                        if ty.as_ref().map_or(false, |r| !r.is_nullable()) {
+                            bail!("no default value exists for type `{ty}`")
+                        }
+                    }
+
                     self.func_new(
                         import.module(),
                         import.name(),
@@ -332,8 +296,10 @@ impl<T> Linker<T> {
                                     ValType::F32 => Val::F32(0.0_f32.to_bits()),
                                     ValType::F64 => Val::F64(0.0_f64.to_bits()),
                                     ValType::V128 => Val::V128(0_u128.into()),
-                                    ValType::FuncRef => Val::FuncRef(None),
-                                    ValType::ExternRef => Val::ExternRef(None),
+                                    ValType::Ref(r) => {
+                                        debug_assert!(r.is_nullable());
+                                        Val::null_ref(r.heap_type())
+                                    }
                                 };
                             }
                             Ok(())
@@ -415,8 +381,11 @@ impl<T> Linker<T> {
     /// Creates a [`Func::new`]-style function named in this linker.
     ///
     /// For more information see [`Linker::func_wrap`].
-    #[cfg(any(feature = "cranelift", feature = "winch"))]
-    #[cfg_attr(nightlydoc, doc(cfg(any(feature = "cranelift", feature = "winch"))))]
+    ///
+    /// # Panics
+    ///
+    /// Panics if the given function type is not associated with the same engine
+    /// as this linker.
     pub fn func_new(
         &mut self,
         module: &str,
@@ -424,6 +393,7 @@ impl<T> Linker<T> {
         ty: FuncType,
         func: impl Fn(Caller<'_, T>, &[Val], &mut [Val]) -> Result<()> + Send + Sync + 'static,
     ) -> Result<&mut Self> {
+        assert!(ty.comes_from_same_engine(self.engine()));
         let func = HostFunc::new(&self.engine, ty, func);
         let key = self.import_key(module, Some(name));
         self.insert(key, Definition::HostFunc(Arc::new(func)))?;
@@ -433,8 +403,11 @@ impl<T> Linker<T> {
     /// Creates a [`Func::new_unchecked`]-style function named in this linker.
     ///
     /// For more information see [`Linker::func_wrap`].
-    #[cfg(any(feature = "cranelift", feature = "winch"))]
-    #[cfg_attr(nightlydoc, doc(cfg(any(feature = "cranelift", feature = "winch"))))]
+    ///
+    /// # Panics
+    ///
+    /// Panics if the given function type is not associated with the same engine
+    /// as this linker.
     pub unsafe fn func_new_unchecked(
         &mut self,
         module: &str,
@@ -442,6 +415,7 @@ impl<T> Linker<T> {
         ty: FuncType,
         func: impl Fn(Caller<'_, T>, &mut [ValRaw]) -> Result<()> + Send + Sync + 'static,
     ) -> Result<&mut Self> {
+        assert!(ty.comes_from_same_engine(self.engine()));
         let func = HostFunc::new_unchecked(&self.engine, ty, func);
         let key = self.import_key(module, Some(name));
         self.insert(key, Definition::HostFunc(Arc::new(func)))?;
@@ -451,8 +425,17 @@ impl<T> Linker<T> {
     /// Creates a [`Func::new_async`]-style function named in this linker.
     ///
     /// For more information see [`Linker::func_wrap`].
+    ///
+    /// # Panics
+    ///
+    /// This method panics in the following situations:
+    ///
+    /// * This linker is not associated with an [async
+    ///   config](crate::Config::async_support).
+    ///
+    /// * If the given function type is not associated with the same engine as
+    ///   this linker.
     #[cfg(all(feature = "async", feature = "cranelift"))]
-    #[cfg_attr(nightlydoc, doc(cfg(all(feature = "async", feature = "cranelift"))))]
     pub fn func_new_async<F>(
         &mut self,
         module: &str,
@@ -474,6 +457,7 @@ impl<T> Linker<T> {
             self.engine.config().async_support,
             "cannot use `func_new_async` without enabling async support in the config"
         );
+        assert!(ty.comes_from_same_engine(self.engine()));
         self.func_new(module, name, ty, move |mut caller, params, results| {
             let async_cx = caller
                 .store
@@ -557,7 +541,44 @@ impl<T> Linker<T> {
         Ok(self)
     }
 
-    for_each_function_signature!(generate_wrap_async_func);
+    /// Asynchronous analog of [`Linker::func_wrap`].
+    #[cfg(feature = "async")]
+    pub fn func_wrap_async<F, Params: WasmTyList, Args: crate::WasmRet>(
+        &mut self,
+        module: &str,
+        name: &str,
+        func: F,
+    ) -> Result<&mut Self>
+    where
+        F: for<'a> Fn(Caller<'a, T>, Params) -> Box<dyn Future<Output = Args> + Send + 'a>
+            + Send
+            + Sync
+            + 'static,
+    {
+        assert!(
+            self.engine.config().async_support,
+            "cannot use `func_wrap_async` without enabling async support on the config",
+        );
+        let func = HostFunc::wrap_inner(
+            &self.engine,
+            move |mut caller: Caller<'_, T>, args: Params| {
+                let async_cx = caller
+                    .store
+                    .as_context_mut()
+                    .0
+                    .async_cx()
+                    .expect("Attempt to start async function on dying fiber");
+                let mut future = Pin::from(func(caller, args));
+                match unsafe { async_cx.block_on(future.as_mut()) } {
+                    Ok(ret) => ret.into_fallible(),
+                    Err(e) => Args::fallible_from_error(e),
+                }
+            },
+        );
+        let key = self.import_key(module, Some(name));
+        self.insert(key, Definition::HostFunc(Arc::new(func)))?;
+        Ok(self)
+    }
 
     /// Convenience wrapper to define an entire [`Instance`] in this linker.
     ///
@@ -753,8 +774,6 @@ impl<T> Linker<T> {
     /// # Ok(())
     /// # }
     /// ```
-    #[cfg(any(feature = "cranelift", feature = "winch"))]
-    #[cfg_attr(nightlydoc, doc(cfg(any(feature = "cranelift", feature = "winch"))))]
     pub fn module(
         &mut self,
         mut store: impl AsContextMut<Data = T>,
@@ -825,7 +844,6 @@ impl<T> Linker<T> {
     ///
     /// This is the same as [`Linker::module`], except for async `Store`s.
     #[cfg(all(feature = "async", feature = "cranelift"))]
-    #[cfg_attr(nightlydoc, doc(cfg(all(feature = "async", feature = "cranelift"))))]
     pub async fn module_async(
         &mut self,
         mut store: impl AsContextMut<Data = T>,
@@ -1098,7 +1116,6 @@ impl<T> Linker<T> {
     /// Attempts to instantiate the `module` provided. This is the same as
     /// [`Linker::instantiate`], except for async `Store`s.
     #[cfg(feature = "async")]
-    #[cfg_attr(nightlydoc, doc(cfg(feature = "async")))]
     pub async fn instantiate_async(
         &self,
         mut store: impl AsContextMut<Data = T>,
@@ -1182,7 +1199,8 @@ impl<T> Linker<T> {
         let mut imports = module
             .imports()
             .map(|import| self._get_by_import(&import))
-            .collect::<Result<Vec<_>, _>>()?;
+            .collect::<Result<Vec<_>, _>>()
+            .err2anyhow()?;
         if let Some(store) = store {
             for import in imports.iter_mut() {
                 import.update_size(store);
@@ -1292,7 +1310,7 @@ impl<T> Linker<T> {
     ) -> Result<Func> {
         if let Some(external) = self.get(&mut store, module, "") {
             if let Extern::Func(func) = external {
-                return Ok(func.clone());
+                return Ok(func);
             }
             bail!("default export in '{}' is not a function", module);
         }
@@ -1300,7 +1318,7 @@ impl<T> Linker<T> {
         // For compatibility, also recognize "_start".
         if let Some(external) = self.get(&mut store, module, "_start") {
             if let Extern::Func(func) = external {
-                return Ok(func.clone());
+                return Ok(func);
             }
             bail!("`_start` in '{}' is not a function", module);
         }
@@ -1366,7 +1384,7 @@ impl DefinitionType {
     pub(crate) fn from(store: &StoreOpaque, item: &Extern) -> DefinitionType {
         let data = store.store_data();
         match item {
-            Extern::Func(f) => DefinitionType::Func(f.sig_index(data)),
+            Extern::Func(f) => DefinitionType::Func(f.type_index(data)),
             Extern::Table(t) => DefinitionType::Table(*t.wasmtime_ty(data), t.internal_size(store)),
             Extern::Global(t) => DefinitionType::Global(*t.wasmtime_ty(data)),
             Extern::Memory(t) => {
@@ -1466,8 +1484,8 @@ impl UnknownImportError {
     }
 }
 
-impl std::fmt::Display for UnknownImportError {
-    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
+impl fmt::Display for UnknownImportError {
+    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
         write!(
             f,
             "unknown import: `{}::{}` has not been defined",
@@ -1476,4 +1494,5 @@ impl std::fmt::Display for UnknownImportError {
     }
 }
 
+#[cfg(feature = "std")]
 impl std::error::Error for UnknownImportError {}
