@@ -1,4 +1,6 @@
 use crate::component::*;
+use crate::prelude::*;
+use crate::Module;
 use crate::ScopeVec;
 use crate::{
     EntityIndex, ModuleEnvironment, ModuleTranslation, ModuleTypesBuilder, PrimaryMap, Tunables,
@@ -10,10 +12,12 @@ use indexmap::IndexMap;
 use std::collections::HashMap;
 use std::mem;
 use wasmparser::types::{
-    AliasableResourceId, ComponentEntityType, ComponentFuncTypeId, ComponentInstanceTypeId, Types,
+    AliasableResourceId, ComponentCoreModuleTypeId, ComponentEntityType, ComponentFuncTypeId,
+    ComponentInstanceTypeId, Types,
 };
 use wasmparser::{Chunk, ComponentImportName, Encoding, Parser, Payload, Validator};
 use wasmtime_types::ModuleInternedTypeIndex;
+use wasmtime_types::WasmResult;
 
 mod adapt;
 pub use self::adapt::*;
@@ -187,7 +191,7 @@ enum LocalInitializer<'data> {
     ResourceDrop(AliasableResourceId, ModuleInternedTypeIndex),
 
     // core wasm modules
-    ModuleStatic(StaticModuleIndex),
+    ModuleStatic(StaticModuleIndex, ComponentCoreModuleTypeId),
 
     // core wasm module instances
     ModuleInstantiate(ModuleIndex, HashMap<&'data str, ModuleInstanceIndex>),
@@ -269,12 +273,14 @@ impl<'a, 'data> Translator<'a, 'data> {
         types: &'a mut ComponentTypesBuilder,
         scope_vec: &'data ScopeVec<u8>,
     ) -> Self {
+        let mut parser = Parser::new(0);
+        parser.set_features(*validator.features());
         Self {
             result: Translation::default(),
             tunables,
             validator,
             types: PreInliningComponentTypes::new(types),
-            parser: Parser::new(0),
+            parser,
             lexical_scopes: Vec::new(),
             static_components: Default::default(),
             static_modules: Default::default(),
@@ -353,7 +359,9 @@ impl<'a, 'data> Translator<'a, 'data> {
             &self.static_components,
         )?;
         self.partition_adapter_modules(&mut component);
-        Ok((component.finish(), self.static_modules))
+        let translation =
+            component.finish(self.types.types_mut_for_inlining(), self.result.types_ref())?;
+        Ok((translation, self.static_modules))
     }
 
     fn translate_payload(
@@ -487,7 +495,7 @@ impl<'a, 'data> Translator<'a, 'data> {
                             let lower_ty = types.component_function_at(func_index);
                             let func = ComponentFuncIndex::from_u32(func_index);
                             let options = self.canonical_options(&options);
-                            let canonical_abi = self.core_func_signature(core_func_index);
+                            let canonical_abi = self.core_func_signature(core_func_index)?;
 
                             core_func_index += 1;
                             LocalInitializer::Lower {
@@ -499,19 +507,19 @@ impl<'a, 'data> Translator<'a, 'data> {
                         }
                         wasmparser::CanonicalFunction::ResourceNew { resource } => {
                             let resource = types.component_any_type_at(resource).unwrap_resource();
-                            let ty = self.core_func_signature(core_func_index);
+                            let ty = self.core_func_signature(core_func_index)?;
                             core_func_index += 1;
                             LocalInitializer::ResourceNew(resource, ty)
                         }
                         wasmparser::CanonicalFunction::ResourceDrop { resource } => {
                             let resource = types.component_any_type_at(resource).unwrap_resource();
-                            let ty = self.core_func_signature(core_func_index);
+                            let ty = self.core_func_signature(core_func_index)?;
                             core_func_index += 1;
                             LocalInitializer::ResourceDrop(resource, ty)
                         }
                         wasmparser::CanonicalFunction::ResourceRep { resource } => {
                             let resource = types.component_any_type_at(resource).unwrap_resource();
-                            let ty = self.core_func_signature(core_func_index);
+                            let ty = self.core_func_signature(core_func_index)?;
                             core_func_index += 1;
                             LocalInitializer::ResourceRep(resource, ty)
                         }
@@ -528,8 +536,12 @@ impl<'a, 'data> Translator<'a, 'data> {
             // Note that this is just initial type translation of the core wasm
             // module and actual function compilation is deferred until this
             // entire process has completed.
-            Payload::ModuleSection { parser, range } => {
-                self.validator.module_section(&range)?;
+            Payload::ModuleSection {
+                parser,
+                unchecked_range,
+            } => {
+                let index = self.validator.types(0).unwrap().module_count();
+                self.validator.module_section(&unchecked_range)?;
                 let translation = ModuleEnvironment::new(
                     self.tunables,
                     self.validator,
@@ -537,21 +549,25 @@ impl<'a, 'data> Translator<'a, 'data> {
                 )
                 .translate(
                     parser,
-                    component.get(range.start..range.end).ok_or_else(|| {
-                        anyhow!(
-                            "section range {}..{} is out of bounds (bound = {})",
-                            range.start,
-                            range.end,
-                            component.len()
-                        )
-                        .context("wasm component contains an invalid module section")
-                    })?,
+                    component
+                        .get(unchecked_range.start..unchecked_range.end)
+                        .ok_or_else(|| {
+                            anyhow!(
+                                "section range {}..{} is out of bounds (bound = {})",
+                                unchecked_range.start,
+                                unchecked_range.end,
+                                component.len()
+                            )
+                            .context("wasm component contains an invalid module section")
+                        })?,
                 )?;
                 let static_idx = self.static_modules.push(translation);
+                let types = self.validator.types(0).unwrap();
+                let ty = types.module_at(index);
                 self.result
                     .initializers
-                    .push(LocalInitializer::ModuleStatic(static_idx));
-                return Ok(Action::Skip(range.end - range.start));
+                    .push(LocalInitializer::ModuleStatic(static_idx, ty));
+                return Ok(Action::Skip(unchecked_range.end - unchecked_range.start));
             }
 
             // When a sub-component is found then the current translation state
@@ -562,8 +578,11 @@ impl<'a, 'data> Translator<'a, 'data> {
             // starts empty since it will only get populated if translation of
             // the nested component ends up aliasing some outer module or
             // component.
-            Payload::ComponentSection { parser, range } => {
-                self.validator.component_section(&range)?;
+            Payload::ComponentSection {
+                parser,
+                unchecked_range,
+            } => {
+                self.validator.component_section(&unchecked_range)?;
                 self.lexical_scopes.push(LexicalScope {
                     parser: mem::replace(&mut self.parser, parser),
                     translation: mem::take(&mut self.result),
@@ -903,12 +922,14 @@ impl<'a, 'data> Translator<'a, 'data> {
         return ret;
     }
 
-    fn core_func_signature(&mut self, idx: u32) -> ModuleInternedTypeIndex {
+    /// Get the interned type index for the `index`th core function.
+    fn core_func_signature(&mut self, index: u32) -> WasmResult<ModuleInternedTypeIndex> {
         let types = self.validator.types(0).unwrap();
-        let id = types.core_function_at(idx);
-        let ty = types[id].unwrap_func();
-        let ty = self.types.convert_func_type(ty);
-        self.types.module_types_builder().wasm_func_type(id, ty)
+        let id = types.core_function_at(index);
+        let module = Module::default();
+        self.types
+            .module_types_builder()
+            .intern_type(&module, types, id)
     }
 }
 
@@ -960,6 +981,13 @@ mod pre_inlining {
         fn lookup_heap_type(&self, index: wasmparser::UnpackedIndex) -> WasmHeapType {
             self.types.lookup_heap_type(index)
         }
+
+        fn lookup_type_index(
+            &self,
+            index: wasmparser::UnpackedIndex,
+        ) -> wasmtime_types::EngineOrModuleTypeIndex {
+            self.types.lookup_type_index(index)
+        }
     }
 }
 use pre_inlining::PreInliningComponentTypes;
